# Copyright 2018-2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the :class:`~pennylane.data.Dataset` class and its associated functions.
"""

from abc import ABC
from glob import glob
import os

from pennylane import Hamiltonian
from pennylane.pauli import string_to_pauli_word, pauli_word_to_string

condensed_hamiltonians = {"hamiltonian", "tapered_hamiltonian"}


def _import_zstd_dill():
    """Import zstd and dill."""
    try:
        # pylint: disable=import-outside-toplevel, unused-import, multiple-imports
        import zstd, dill
    except ImportError as Error:
        raise ImportError(
            "This feature requires zstd and dill. "
            "They can be installed with:\n\n pip install zstd dill."
        ) from Error

    return zstd, dill


class DatasetLoadError(Exception):
    """Error raised when a Dataset has trouble finding a lazy-loaded attribute."""


# pylint: disable=too-many-instance-attributes
class Dataset(ABC):
    """Create a dataset object to store a collection of information describing
    a physical system and its evolution. For example, a dataset for an arbitrary
    quantum system could have a Hamiltonian, its ground state,
    and an efficient state-preparation circuit for that state.

    Args:
        *args: For internal use only. These will be ignored if called with ``standard=False``
        standard (bool): For internal use only. See the note below for the behavior when this is set to ``True``
        **kwargs: variable-length keyword arguments specifying the data to be stored in the dataset

    Note on the ``standard`` kwarg:
        A `standard` Dataset uses previously generated, hosted quantum data. This special instance of the
        ``Dataset`` class makes certain assumptions about folder management for downloading the data and
        handling I/O. As such, the ``Dataset`` class should not be instantiated by the users directly with
        ``standard=True``. Instead, they should use :meth:`~load`.


    .. seealso:: :meth:`~load`

    **Example**

    >>> Hamiltonian = qml.Hamiltonian([1., 1.], [qml.PauliZ(wires=0), qml.PauliZ(wires=1)])
    >>> eigvals, eigvecs = np.linalg.eigh(qml.matrix(Hamiltonian))
    >>> ground_state_energy = np.min(eigvals)
    >>> ground_state = np.transpose(eigvecs)[np.argmin(eigvals)]
    >>> dataset = qml.data.Dataset(Hamiltonian = Hamiltonian, ground_state = ground_state,
            ground_state_energy = ground_state_energy)
    >>> print(dataset.Hamiltonian)
          (1) [Z0]
        + (1) [Z1]
    >>> print(dataset.ground_energy)
    -2.0

    .. details::
        :title: Usage Details

        In addition to creating datasets in memory, we can also store them in
        the disk and then load them as follows. First we create the dataset:

        >>> Hamiltonian = qml.Hamiltonian([1., 1.], [qml.PauliZ(wires=0), qml.PauliZ(wires=1)])
        ...     eigvals, eigvecs = np.linalg.eigh(qml.matrix(Hamiltonian))
        >>> ground_state_energy = np.min(eigvals)
        >>> ground_state = np.transpose(eigvecs)[np.argmin(eigvals)]
        >>> dataset = qml.data.Dataset(Hamiltonian = Hamiltonian, ground_state = ground_state,
        ...     ground_state_energy = ground_state_energy)

        Then to save the dataset to a file, we call :func:`Dataset.write()`:

        >>> dataset.write('./path/to/file/dataset.dat')

        We can then retrieve the data using :func:`Dataset.read()`

        >>> retrieved_data = qml.data.Dataset()
        >>> retrieved_data.read('./path/to/file/dataset.dat')
        >>> print(retrieved_data.Hamiltonian)
          (1) [Z0]
        + (1) [Z1]
        >>> print(dataset.Hamiltonian)
          (1) [Z0]
        + (1) [Z1]
    """

    _standard_argnames = ["data_name", "data_folder", "attr_prefix", "docstring"]

    def __std_init__(self, data_name, folder, attr_prefix, docstring):
        """Constructor for standardized datasets."""
        self._dtype = data_name
        self._folder = folder.rstrip(os.path.sep)
        self._description = os.path.join(data_name, self._folder.split(data_name)[-1][1:])
        self.__doc__ = docstring

        prefix = os.path.join(self._folder, attr_prefix) + "_{}.dat"
        self._fullfile = prefix.format("full")
        if os.path.exists(self._fullfile):
            self.read(self._fullfile, lazy=True)
        else:
            self._fullfile = None
            prefix_len = len(attr_prefix) + 1
            for f in glob(prefix.format("*")):
                attribute = os.path.basename(f)[prefix_len:-4]
                setattr(self, f"{attribute}", None)
                self._attr_filemap[attribute] = (f, False)

    def __base_init__(self, **kwargs):
        """Constructor for user-defined datasets."""
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __init__(self, *args, standard=False, **kwargs):
        """Dispatching constructor for direct invocation with kwargs or from qml.data.load()"""
        self._attr_filemap = {}
        if standard:
            if len(args) != len(self._standard_argnames):
                raise TypeError(
                    f"Standard datasets expect {len(self._standard_argnames)} arguments: {self._standard_argnames}"
                )
            for name, val in zip(self._standard_argnames, args):
                if not isinstance(val, str):
                    raise ValueError(f"Expected {name} to be a str, got {type(val).__name__}")
            self._is_standard = True
            self.__std_init__(*args)
        else:
            self._is_standard = False
            self.__base_init__(**kwargs)

    def __repr__(self):
        attr_str = (
            str(list(self.attrs))
            if len(self.attrs) < 3
            else f"{str(list(self.attrs)[:2])[:-1]}, ...]"
        )

        std_str = f"description: {self._description}, " if self._is_standard else ""
        return f"<Dataset = {std_str}attributes: {attr_str}>"

    @property
    def attrs(self):
        """Returns attributes of the dataset."""
        return {k: v for k, v in vars(self).items() if k[0] != "_"}

    # pylint:disable=c-extension-no-member
    @staticmethod
    def _read_file(filepath):
        """Reads data from a previously created or downloaded data file.

        Returns:
            A dictionary for non-standard datasets or full files, otherwise a data value
        """
        with open(filepath, "rb") as f:
            compressed_pickle = f.read()

        zstd, dill = _import_zstd_dill()
        depressed_pickle = zstd.decompress(compressed_pickle)
        return dill.loads(depressed_pickle)

    def read(self, filepath, lazy=False, assign_to=None):
        """Loads data from a saved file to the current dataset.

        Args:
            filepath (string): The desired location and filename to load, e.g. './path/to/file/file_name.dat'.
            lazy (bool): Indicates if only the key of the attribute should be saved to the Dataset instance.
                Note that the file will be remembered and its contents will be loaded when the attribute is used.
            assign_to (str): Attribute name to which the contents of the file should be assigned.
                If this is ``None`` (the default value), this method will assume that the file contents are of
                the form ``{attribute_name: attribute_value,}``.

        **Example**

        >>> new_dataset = qml.data.Dataset(kw1 = 1, kw2 = '2', kw3 = [3])
        >>> new_dataset.read('./path/to/file/file_name.dat')

        Using the ``assign_to`` keyword argument:

        >>> new_dataset = qml.data.Dataset()
        >>> new_dataset.read('./path/to/file/single_state.dat', assign_to="state")
        >>> new_dataset.state  # assuming the above file contains only a tensor
        tensor([1, 1, 0, 0], requires_grad=True)
        """
        data = self._read_file(filepath)
        file_contains_dataset = True
        if assign_to is not None:
            data = {assign_to: data}
            file_contains_dataset = False
        if lazy:
            for attr in data:
                setattr(self, f"{attr}", None)
                self._attr_filemap[attr] = (filepath, file_contains_dataset)
            return
        for attr, value in data.items():  # pylint:disable=no-member
            if attr in condensed_hamiltonians:
                value = self.__dict_to_hamiltonian(value["terms"], value["wire_map"])
            setattr(self, f"{attr}", value)

    # pylint:disable=c-extension-no-member
    @staticmethod
    def _write_file(data, filepath, protocol=4):
        """General method to write data to a file."""
        zstd, dill = _import_zstd_dill()
        pickled_data = dill.dumps(data, protocol=protocol)
        compressed_pickle = zstd.compress(pickled_data)
        with open(filepath, "wb") as f:
            f.write(compressed_pickle)

    def write(self, filepath, protocol=4):
        """Writes the dataset to a file as a dictionary.

        Args:
            filepath (string): the desired save location and file name

        **Example**

        >>> new_dataset = qml.data.Dataset(kw1 = 1, kw2 = '2', kw3 = [3])
        >>> new_dataset.write('./path/to/file/file_name.dat')
        """
        dirname = os.path.dirname(filepath)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        # if some values are still lazy-loaded to None, this will load them before writing
        if self._is_standard:
            for attr in self.attrs:
                _ = getattr(self, attr)
        attrs = self.attrs
        for h in condensed_hamiltonians.intersection(attrs):
            attrs[h] = self.__hamiltonian_to_dict(attrs[h])
        self._write_file(attrs, filepath, protocol=protocol)

    def list_attributes(self):
        """List the attributes saved on the Dataset"""
        return list(self.attrs)

    # pylint: disable=attribute-defined-outside-init
    def __copy__(self):
        cls = self.__class__

        if not self._is_standard:
            copied = cls(**self.attrs)
            copied._attr_filemap = self._attr_filemap.copy()
            return copied

        copied = cls.__new__(cls)
        copied._is_standard = True
        copied._dtype = self._dtype
        copied._folder = self._folder
        copied._fullfile = self._fullfile
        copied._description = self._description
        copied._attr_filemap = self._attr_filemap.copy()
        copied.__doc__ = self.__doc__
        for key, val in self.attrs.items():
            setattr(copied, f"{key}", val)

        return copied

    def __getattribute__(self, name):
        value = super().__getattribute__(name)
        if value is not None or name not in self._attr_filemap:
            return value

        filepath, file_contains_dataset = self._attr_filemap[name]
        if not os.path.exists(filepath):
            raise DatasetLoadError(
                f"Dataset lazy-loaded a '{name}' attribute, but the file originally containing it ({filepath}) was not found."
            )

        value = self._read_file(filepath)
        if file_contains_dataset:
            if name not in value:
                raise DatasetLoadError(
                    f"Dataset lazy-loaded a '{name}' attribute from {filepath}, but it no longer appears to be in the file."
                )
            value = value[name]
        if name in condensed_hamiltonians:
            value = self.__dict_to_hamiltonian(value["terms"], value["wire_map"])

        setattr(self, name, value)
        del self._attr_filemap[name]
        return value

    @staticmethod
    def __dict_to_hamiltonian(terms, wire_map):
        """Converts a dict of terms and a wire map into a Hamiltonian instance."""
        coeffs, ops = [], []
        for pauli_string, coeff in terms.items():
            coeffs.append(coeff)
            ops.append(string_to_pauli_word(pauli_string, wire_map))
        return Hamiltonian(coeffs, ops)

    @staticmethod
    def __hamiltonian_to_dict(hamiltonian):
        """Converts a hamiltonian instance into a dict containing pauli-string terms and a wire map."""
        coeffs, ops = hamiltonian.terms()
        wire_map = {w: i for i, w in enumerate(hamiltonian.wires)}
        return {
            "terms": {pauli_word_to_string(op, wire_map): coeff for coeff, op in zip(coeffs, ops)},
            "wire_map": wire_map,
        }



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains utilities and auxiliary functions which are shared
across the PennyLane submodules.
"""
# pylint: disable=protected-access,too-many-branches
from collections.abc import Iterable
import functools
import inspect
import numbers
import warnings


import numpy as np

import pennylane as qml


def sparse_hamiltonian(H, wires=None):
    r"""Warning: This method is deprecated. Use :meth:~.Hamiltonian.sparse_matrix` instead.

    Computes the sparse matrix representation a Hamiltonian in the computational basis.

    Args:
        H (~.Hamiltonian): Hamiltonian operator for which the matrix representation should be
            computed
        wires (Iterable): Wire labels that indicate the order of wires according to which the matrix
            is constructed. If not profided, ``H.wires`` is used.

    Returns:
        csr_matrix: a sparse matrix in scipy Compressed Sparse Row (CSR) format with dimension
        :math:`(2^n, 2^n)`, where :math:`n` is the number of wires

    **Example:**

    This function can be used by passing a `qml.Hamiltonian` object as:

    >>> coeffs = [1, -0.45]
    >>> obs = [qml.PauliZ(0) @ qml.PauliZ(1), qml.PauliY(0) @ qml.PauliZ(1)]
    >>> H = qml.Hamiltonian(coeffs, obs)
    >>> H_sparse = sparse_hamiltonian(H)
    >>> H_sparse
    <4x4 sparse matrix of type '<class 'numpy.complex128'>'
        with 2 stored elements in COOrdinate format>

    The resulting sparse matrix can be either used directly or transformed into a numpy array:

    >>> H_sparse.toarray()
    array([[ 1.+0.j  ,  0.+0.j  ,  0.+0.45j,  0.+0.j  ],
           [ 0.+0.j  , -1.+0.j  ,  0.+0.j  ,  0.-0.45j],
           [ 0.-0.45j,  0.+0.j  , -1.+0.j  ,  0.+0.j  ],
           [ 0.+0.j  ,  0.+0.45j,  0.+0.j  ,  1.+0.j  ]])

    """
    warnings.warn(
        "The method sparse_hamiltonian is deprecated. Please use the method "
        "sparse_matrix of the Hamiltonian operator instead.",
        UserWarning,
    )
    if not isinstance(H, qml.Hamiltonian):
        raise TypeError("Passed Hamiltonian must be of type `qml.Hamiltonian`")

    if wires is None:  # not sure if this if-else is still necessary
        wires = H.wires
    else:
        wires = qml.wires.Wires(wires)

    return H.sparse_matrix(wire_order=wires)


def _flatten(x):
    """Iterate recursively through an arbitrarily nested structure in depth-first order.

    See also :func:`_unflatten`.

    Args:
        x (array, Iterable, Any): each element of an array or an Iterable may itself be any of these types

    Yields:
        Any: elements of x in depth-first order
    """
    if isinstance(x, np.ndarray):
        yield from _flatten(x.flat)  # should we allow object arrays? or just "yield from x.flat"?
    elif isinstance(x, qml.wires.Wires):
        # Reursive calls to flatten `Wires` will cause infinite recursion (`Wires` atoms are `Wires`).
        # Since Wires are always flat, just yield.
        for item in x:
            yield item
    elif isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
        for item in x:
            yield from _flatten(item)
    else:
        yield x


def _unflatten(flat, model):
    """Restores an arbitrary nested structure to a flattened iterable.

    See also :func:`_flatten`.

    Args:
        flat (array): 1D array of items
        model (array, Iterable, Number): model nested structure

    Raises:
        TypeError: if ``model`` contains an object of unsupported type

    Returns:
        Union[array, list, Any], array: first elements of flat arranged into the nested
        structure of model, unused elements of flat
    """
    if isinstance(model, (numbers.Number, str)):
        return flat[0], flat[1:]

    if isinstance(model, np.ndarray):
        idx = model.size
        res = np.array(flat)[:idx].reshape(model.shape)
        return res, flat[idx:]

    if isinstance(model, Iterable):
        res = []
        for x in model:
            val, flat = _unflatten(flat, x)
            res.append(val)
        return res, flat

    raise TypeError(f"Unsupported type in the model: {type(model)}")


def unflatten(flat, model):
    """Wrapper for :func:`_unflatten`.

    Args:
        flat (array): 1D array of items
        model (array, Iterable, Number): model nested structure

    Raises:
        ValueError: if ``flat`` has more elements than ``model``
    """
    # pylint:disable=len-as-condition
    res, tail = _unflatten(np.asarray(flat), model)
    if len(tail) != 0:
        raise ValueError("Flattened iterable has more elements than the model.")
    return res


def _inv_dict(d):
    """Reverse a dictionary mapping.

    Returns multimap where the keys are the former values,
    and values are sets of the former keys.

    Args:
        d (dict[a->b]): mapping to reverse

    Returns:
        dict[b->set[a]]: reversed mapping
    """
    ret = {}
    for k, v in d.items():
        ret.setdefault(v, set()).add(k)
    return ret


def _get_default_args(func):
    """Get the default arguments of a function.

    Args:
        func (callable): a function

    Returns:
        dict[str, tuple]: mapping from argument name to (positional idx, default value)
    """
    signature = inspect.signature(func)
    return {
        k: (idx, v.default)
        for idx, (k, v) in enumerate(signature.parameters.items())
        if v.default is not inspect.Parameter.empty
    }


@functools.lru_cache()
def pauli_eigs(n):
    r"""Eigenvalues for :math:`A^{\otimes n}`, where :math:`A` is
    Pauli operator, or shares its eigenvalues.

    As an example if n==2, then the eigenvalues of a tensor product consisting
    of two matrices sharing the eigenvalues with Pauli matrices is returned.

    Args:
        n (int): the number of qubits the matrix acts on
    Returns:
        list: the eigenvalues of the specified observable
    """
    if n == 1:
        return np.array([1, -1])
    return np.concatenate([pauli_eigs(n - 1), -pauli_eigs(n - 1)])


def expand_vector(vector, original_wires, expanded_wires):
    r"""Expand a vector to more wires.

    Args:
        vector (array): :math:`2^n` vector where n = len(original_wires).
        original_wires (Sequence[int]): original wires of vector
        expanded_wires (Union[Sequence[int], int]): expanded wires of vector, can be shuffled
            If a single int m is given, corresponds to list(range(m))

    Returns:
        array: :math:`2^m` vector where m = len(expanded_wires).
    """
    if isinstance(expanded_wires, numbers.Integral):
        expanded_wires = list(range(expanded_wires))

    N = len(original_wires)
    M = len(expanded_wires)
    D = M - N

    if not set(expanded_wires).issuperset(original_wires):
        raise ValueError("Invalid target subsystems provided in 'original_wires' argument.")

    if qml.math.shape(vector) != (2**N,):
        raise ValueError("Vector parameter must be of length 2**len(original_wires)")

    dims = [2] * N
    tensor = qml.math.reshape(vector, dims)

    if D > 0:
        extra_dims = [2] * D
        ones = qml.math.ones(2**D).reshape(extra_dims)
        expanded_tensor = qml.math.tensordot(tensor, ones, axes=0)
    else:
        expanded_tensor = tensor

    wire_indices = []
    for wire in original_wires:
        wire_indices.append(expanded_wires.index(wire))

    wire_indices = np.array(wire_indices)

    # Order tensor factors according to wires
    original_indices = np.array(range(N))
    expanded_tensor = qml.math.moveaxis(
        expanded_tensor, tuple(original_indices), tuple(wire_indices)
    )

    return qml.math.reshape(expanded_tensor, 2**M)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains functions for computing the stochastic parameter-shift gradient
of pulse sequences in a qubit-based quantum tape.
"""
from collections.abc import Sequence
import numpy as np

import pennylane as qml
from pennylane._device import _get_num_copies
from pennylane.measurements import MutualInfoMP, StateMP, VarianceMP, VnEntropyMP
from pennylane.pulse import ParametrizedEvolution

from .finite_difference import _all_zero_grad_new, _no_trainable_grad_new
from .parameter_shift import (
    _reorder_grad_axes_multi_measure,
    _reorder_grad_axes_single_measure_shot_vector,
    _make_zero_rep,
)
from .gradient_transform import (
    choose_grad_methods,
    grad_method_validation,
    gradient_analysis,
    gradient_transform,
)

has_jax = True
try:
    import jax
    import jax.numpy as jnp
except ImportError:
    has_jax = False


def _split_evol_ops(op, word, word_wires, tau):
    r"""Randomly split a ``ParametrizedEvolution`` with respect to time into two operations and
    insert a Pauli rotation using a given Pauli word and rotation angles :math:`\pm\pi/2`.
    This yields two groups of three operations each.

    Args:
        op (ParametrizedEvolution): operation to split up.
        word (str): Pauli word with respect to which to rotate between the split up parts.
        word_wires (~.wires.Wires): wires on which the Pauli word acts
        tau (float or tensor_like): split-up time(s). If multiple times are passed, the split-up
            operations are set up to return intermediate time evolution results, leading to
            broadcasting effectively.

    Returns:
        tuple[list[`~.Operation`]]: The split-time evolution, expressed as three operations in the
            inner lists. The first tuple entry contains the operations with positive Pauli rotation
            angle, the second entry the operations with negative Pauli rotation angle.
    """
    t0, *_, t1 = op.t
    if broadcast := qml.math.ndim(tau) > 0:
        tau = jnp.sort(tau)
        before_t = jnp.concatenate([jnp.array([t0]), tau, jnp.array([t1])])
        after_t = before_t.copy()
    else:
        before_t = jax.numpy.array([t0, tau])
        after_t = jax.numpy.array([tau, t1])

    before_plus, before_minus = (
        op(op.data, before_t, return_intermediate=broadcast, **op.odeint_kwargs) for _ in range(2)
    )
    after_plus, after_minus = (
        op(
            op.data,
            after_t,
            return_intermediate=broadcast,
            complementary=broadcast,
            **op.odeint_kwargs,
        )
        for _ in range(2)
    )
    # Create Pauli rotations to be inserted at tau
    evolve_plus = qml.PauliRot(np.pi / 2, word, wires=word_wires)
    evolve_minus = qml.PauliRot(-np.pi / 2, word, wires=word_wires)
    # Construct gate sequences of split intervals and inserted Pauli rotations
    ops = ([before_plus, evolve_plus, after_plus], [before_minus, evolve_minus, after_minus])
    return ops


def _split_evol_tapes(tape, split_evolve_ops, op_idx):
    """Replace a marked ``ParametrizedEvolution`` in a given tape by provided operations, creating
    one tape per group of operations.

    Args:
        tape (QuantumTape): original tape
        split_evolve_ops (tuple[list[qml.Operation]]): The time-split evolution operations as
            created by ``_split_evol_ops``. For each group of operations, a new tape
            is created.
        op_idx (int): index of the operation to replace within the tape

    Returns:
        list[QuantumTape]: new tapes with replaced operation, one tape per group of operations in
        ``split_evolve_ops``.
    """
    ops_pre = tape.operations[:op_idx]
    ops_post = tape.operations[op_idx + 1 :]
    return [
        qml.tape.QuantumScript(ops_pre + split + ops_post, tape.measurements)
        for split in split_evolve_ops
    ]


# pylint: disable=too-many-arguments
def _parshift_and_integrate(
    results, cjacs, int_prefactor, single_measure, shot_vector, use_broadcasting
):
    """Apply the parameter-shift rule post-processing to tape results and contract
    with classical Jacobians, effectively evaluating the numerical integral of the stochastic
    parameter-shift rule.

    Args:
        results (list): Tape evaluation results, corresponding to the modified quantum
            circuit result when using the applicable parameter shifts and the sample splitting
            times. Results should be ordered such that the different shifted circuits for a given
            splitting time are grouped together
        cjacs (tensor_like): classical Jacobian evaluated at the splitting times
        int_prefactor (float): prefactor of the numerical integration, corresponding to the size
            of the time range divided by the number of splitting time samples
        single_measure (bool): Whether the results contain a single measurement per shot setting
        shot_vector (bool): Whether the results have a shot vector axis
        use_broadcasting (bool): Whether broadcasting was used in the tapes that returned the
            ``results``.
    Returns:
        tensor_like or tuple[tensor_like] or tuple[tuple[tensor_like]]: Gradient entry
    """

    if use_broadcasting:

        def _diff_and_contract(res_list, cjacs, int_prefactor):
            # This one will not work yet with tuples.
            diff = (res := qml.math.stack(res_list))[0, 1:-1] - res[1, 1:-1]
            return qml.math.tensordot(diff, cjacs, axes=[[0], [0]]) * int_prefactor

    else:

        def _diff_and_contract(res_list, cjacs, int_prefactor):
            diff = (res := qml.math.stack(res_list))[::2] - res[1::2]
            return qml.math.tensordot(diff, cjacs, axes=[[0], [0]]) * int_prefactor

    # If multiple measure xor shot_vector: One axis to pull out of the shift rule and integration
    if not single_measure + shot_vector == 1:
        return tuple(_diff_and_contract(r, cjacs, int_prefactor) for r in zip(*results))
    if single_measure:
        # Single measurement without shot vector
        return _diff_and_contract(results, cjacs, int_prefactor)

    # Multiple measurements with shot vector. Not supported with broadcasting yet.
    if use_broadcasting:
        # TODO: Remove once #2690 is resolved
        raise NotImplementedError(
            "Broadcasting, multiple measurements and shot vectors are currently not "
            "supported all simultaneously by stoch_pulse_grad."
        )
    return tuple(
        tuple(_diff_and_contract(_r, cjacs, int_prefactor) for _r in zip(*r)) for r in zip(*results)
    )


# pylint: disable=too-many-arguments
def _stoch_pulse_grad(
    tape, argnum=None, num_split_times=1, sampler_seed=None, shots=None, use_broadcasting=False
):
    r"""Compute the gradient of a quantum circuit composed of pulse sequences by applying the
    stochastic parameter shift rule.

    For a pulse-based cost function :math:`C(\boldsymbol{v}, T)`
    with variational parameters :math:`\boldsymbol{v}` and evolution time :math:`T`, it is given by
    (c.f. Eqn. (6) in `Leng et al. (2022) <https://arxiv.org/abs/2210.15812>`__ with altered
    notation):

    .. math::

        \frac{\partial C}{\partial v_k}
        = \int_{0}^{T} \mathrm{d}\tau \sum_{j=1}^m
        \frac{\partial f_j}{\partial v_k}(\boldsymbol{v}, \tau)
        \left[C_j^{(+)}(\boldsymbol{v}, \tau) - C_j^{(-)}(\boldsymbol{v}, \tau)\right]

    Here, :math:`f_j` are the pulse envelopes that capture the time dependence of the pulse
    Hamiltonian:

    .. math::

        H(\boldsymbol{v}, t) = H_\text{drift} + \sum_j f_j(\boldsymbol{v}, t) H_j,

    and :math:`C_j^{(\pm)}` are modified cost functions:

    .. math::
            C_j^{(\pm)}(\boldsymbol{v}, \tau)&=
            \bra{\psi^{(\pm)}_{j}(\boldsymbol{v}, \tau)} B
            \ket{\psi^{(\pm)}_{j}(\boldsymbol{v}, \tau)} \\
            \ket{\psi^{(\pm)}_{j}(\boldsymbol{v}, \tau)}
            &= U_{\boldsymbol{v}}(T, \tau) e^{-i \pm \frac{\pi}{2} H_j}
            U_{\boldsymbol{v}}(\tau, 0)\ket{\psi_0}.

    That is, the :math:`j`\ th modified time evolution in these circuit interrupts the
    evolution generated by the pulse Hamiltonian by inserting a rotation gate generated by
    the corresponding Hamiltonian term :math:`H_j` with a rotation angle of
    :math:`\pm\frac{\pi}{2}`.

    See below for a more detailed description. The integral in the first equation above
    is estimated numerically in the stochastic parameter-shift rule. For this, it samples
    split times :math:`\tau` and averages the modified cost functions and the Jacobians
    of the envelopes :math:`\partial f_j / \partial v_k` at the sampled times suitably.

    Args:
        tape (pennylane.QNode or .QuantumTape): quantum tape or QNode to differentiate
        argnum (int or list[int] or None): Trainable tape parameter indices to differentiate
            with respect to. If not provided, the derivatives with respect to all
            trainable parameters are returned.
        num_split_times (int): number of time samples to use in the stochastic parameter-shift
            rule underlying the differentiation; also see details
        sample_seed (int): randomness seed to be used for the time samples in the stochastic
            parameter-shift rule
        shots (None, int, list[int]): The shots of the device used to execute the tapes which are
            returned by this transform. Note that this argument does not *influence* the shots
            used for execution, but informs the transform about the shots to ensure a compatible
            return value structure.
        use_broadcasting (bool): Whether to use broadcasting across the different sampled
            splitting times. If ``False`` (the default), one set of modified tapes per
            splitting time is created, if ``True`` only a single set of broadcasted, modified
            tapes is created, increasing performance on simulators.

    Returns:
        function or tuple[list[QuantumTape], function]:

        - If the input is a QNode, an object representing the Jacobian (function) of the QNode
          that can be executed to obtain the Jacobian.
          The type of the Jacobian returned is either a tensor, a tuple or a
          nested tuple depending on the nesting structure of the original QNode output.

        - If the input is a tape, a tuple containing a
          list of generated tapes, together with a post-processing
          function to be applied to the results of the evaluated tapes
          in order to obtain the Jacobian.

    This transform realizes the stochastic parameter-shift rule for pulse sequences, as introduced
    in `Banchi and Crooks (2018) <https://quantum-journal.org/papers/q-2021-01-25-386/>`_ and
    `Leng et al. (2022) <https://arxiv.org/abs/2210.15812>`_.

    .. note::

        This function requires the JAX interface and does not work with other autodiff interfaces
        commonly encountered with PennyLane.
        Finally, this transform is not JIT-compatible yet.

    .. note::

        This function uses a basic sampling approach with a uniform distribution to estimate the
        integral appearing in the stochastic parameter-shift rule. In many cases, there are
        probability distributions that lead to smaller variances of the estimator.
        In addition, the sampling approach will not reduce trivially to simpler parameter-shift
        rules when used with simple pulses (see details and examples below), potentially leading
        to imprecise results and/or unnecessarily large computational efforts.

    .. note::

        Currently this function only supports pulses for which each *parametrized* term is a
        simple Pauli word. More general Hamiltonian terms are not supported yet.

    **Examples**

    Consider a pulse program with a single two-qubit pulse, generated by a Hamiltonian
    with three terms: the non-trainable term :math:`\frac{1}{2}X\otimes I`, the trainable
    constant (over time) term :math:`v_1 Z\otimes Z` and the trainable sinoidal term
    :math:`\sin(v_2 t) I\otimes X`.

    .. code-block:: python

        from pennylane.gradients import stoch_pulse_grad

        dev = qml.device("default.qubit.jax", wires=2)

        def sin(p, t):
            return jax.numpy.sin(p * t)

        ZZ = qml.PauliZ(0) @ qml.PauliZ(1)
        H = 0.5 * qml.PauliX(0) + qml.pulse.constant * ZZ + sin * qml.PauliX(1)

        def ansatz(params):
            qml.evolve(H)(params, (0.2, 0.4))
            return qml.expval(qml.PauliY(1))

        qnode = qml.QNode(ansatz, dev, interface="jax", diff_method=qml.gradients.stoch_pulse_grad)

    The program takes the two parameters :math:`v_1, v_2` for the two trainable terms:

    >>> params = [jax.numpy.array(0.4), jax.numpy.array(1.3)]
    >>> qnode(params)
    Array(-0.15052551, dtype=float32)

    And as we registered the differentiation method :func:`~.stoch_pulse_grad`,
    we can compute its gradient in a hardware compatible manner:

    >>> jax.grad(qnode)(params)
    [Array(0.00277838, dtype=float32, weak_type=True),
     Array(-0.07787319, dtype=float32, weak_type=True)] # results may differ

    Note that the derivative is computed using a stochastic parameter-shift rule,
    which is based on a sampled approximation of an integral expression (see theoretical
    background below). This makes the computed derivative an approximate quantity subject
    to statistical fluctuations with notable variance. The number of samples used to
    approximate the integral can be chosen with ``num_split_times``, the seed for the
    sampling can be fixed with ``sampler_seed``:

    .. code-block:: python

        qnode = qml.QNode(
            ansatz,
            dev,
            interface="jax",
            diff_method=qml.gradients.stoch_pulse_grad,
            num_split_times=5, # Use 5 samples for the approximation
            sampler_seed=18, # Fix randomness seed
        )

    >>> jax.grad(qnode)(params)
    [Array(0.00246266, dtype=float32, weak_type=True),
     Array(-0.11399216, dtype=float32, weak_type=True)]

    On simulator devices, we may activate the option ``use_broadcasting``, which makes
    use of broadcasting internally to improve the performance of the stochastic parameter-shift
    rule:

    .. code-block:: python

        from time import process_time
        faster_grad_qnode = qml.QNode(
            ansatz,
            dev,
            interface="jax",
            diff_method=qml.gradients.stoch_pulse_grad,
            num_split_times=5, # Use 5 samples for the approximation
            sampler_seed=18, # Fix randomness seed
            use_broadcasting=True, # Activate broadcasting
        )
        times = []
        for node in [qnode, faster_grad_qnode]:
            start = process_time()
            jax.grad(node)(params)
            times.append(process_time() - start)

    >>> print(times) # Show the gradient computation times in seconds.
    [11.582010403000002, 3.9708045299999988]

    .. warning::

        As the option ``use_broadcasting=True`` adds a broadcasting dimension to the modified
        circuits, it is not compatible with circuits that already are broadcasted.

    .. details::
        :title: Theoretical background
        :href: theory

        Consider a pulse generated by a time-dependent Hamiltonian

        .. math::

            H(\boldsymbol{v}, t) = H_\text{drift} + \sum_j f_j(v_j, t) H_j,

        where :math:`\boldsymbol{v}=\{v_j\}` are variational parameters and :math:`t` is the time.
        In addition, consider a cost function that is based on using this pulse for
        a duration :math:`T`
        in a pulse sequence and measuring the expectation value of an observable.
        For simplicity we absorb the parts of the sequence
        before and after the considered pulse into the initial state and the observable,
        respectively:

        .. math::

            C(\boldsymbol{v}, t) =
            \bra{\psi_0} U_{\boldsymbol{v}}(T, 0)^\dagger B U_{\boldsymbol{v}}(T, 0)\ket{\psi_0}.

        Here, we denoted the unitary evolution under :math:`H(\boldsymbol{v}, t)` from time
        :math:`t_1` to :math:`t_2` as :math:`U_{\boldsymbol{v}(t_2, t_1)}`.
        Then the derivative of :math:`C` with respect to a specific parameter :math:`v_k`
        is given by (see Eqn. (6) of `Leng et al. (2022) <https://arxiv.org/abs/2210.15812>`_)

        .. math::

            \frac{\partial C}{\partial v_k}
            = \int_{0}^{T} \mathrm{d}\tau \sum_{j=1}^m
            \frac{\partial f_j}{\partial v_k}(\boldsymbol{v}, \tau)
            \widetilde{C_j}(\boldsymbol{v}, \tau).

        Here, the integral ranges over the duration of the pulse, the partial derivatives of
        the coefficient functions, :math:`\partial f_j / \partial v_k`, are computed classically,
        and :math:`\widetilde{C_j}` is a linear combination of the results from modified pulse
        sequence executions based on generalized parameter-shift rules
        (see e.g. `Kyriienko and Elfving (2022) <https://arxiv.org/abs/2108.01218>`_ or
        `Wierichs et al. (2022) <https://doi.org/10.22331/q-2022-03-30-677>`_ for more details
        and :func:`~.param_shift` for an implementation of the non-stochastic generalized shift
        rules)
        Given the parameter shift rule with coefficients :math:`\{y_\ell\}` and shifts
        :math:`\{x_\ell\}` for the single-parameter pulse :math:`\exp(-i \theta H_j)`,
        the linear combination is given by

        .. math::

            \widetilde{C_j}(\boldsymbol{v}, \tau)&=\sum_{\ell=1} y_\ell
            \bra{\psi_{j}(\boldsymbol{v}, x_\ell, \tau)} B
            \ket{\psi_{j}(\boldsymbol{v}, x_\ell, \tau)} \\
            \ket{\psi_{j}(\boldsymbol{v}, x_\ell, \tau)}
            &= U_{\boldsymbol{v}}(T, \tau) e^{-i x_\ell H_j}
            U_{\boldsymbol{v}}(\tau, 0)\ket{\psi_0}.

        In practice, the time integral over :math:`\tau` is computed by sampling values for
        the time, evaluating the integrand, and averaging appropriately. The probability
        distribution used for the sampling may have a significant impact on the quality of the
        obtained estimates, in particular with regards to their variance.
        In this function, a uniform distribution over the interval :math:`[0, t]` is used,
        which often can be improved upon.

        **Examples**

        Consider the pulse generated by

        .. math::

            H(\boldsymbol{v}, t) = \frac{1}{2} X\otimes I + v_1 Z\otimes Z + \sin(v_2 t) I\otimes X

        and the observable :math:`B=I\otimes Y`. There are two variational parameters, :math:`v_1`
        and :math:`v_2`, for which we may compute the derivative of the cost function:

        .. math::

            \frac{\partial C}{\partial v_1}
            &= \int_{0}^{T} \mathrm{d}\tau \ \widetilde{C_1}((v_1, v_2), \tau)\\
            \frac{\partial C}{\partial v_2}
            &= \int_{0}^{T} \mathrm{d}\tau \cos(v_2 \tau) \tau \ \widetilde{C_2}((v_1, v_2), \tau)\\
            \widetilde{C_j}((v_1, v_2), \tau)&=
            \bra{\psi_{j}((v_1, v_2), \pi/4, \tau)} B
            \ket{\psi_{j}((v_1, v_2), \pi/4, \tau)}\\
            &-\bra{\psi_{j}((v_1, v_2), -\pi/4, \tau)} B
            \ket{\psi_{j}((v_1, v_2), -\pi/4, \tau)} \\
            \ket{\psi_{j}((v_1, v_2), x, \tau)}
            &= U_{(v_1, v_2)}(T, \tau) e^{-i x H_j}U_{(v_1, v_2)}(\tau, 0)\ket{0}.

        Here we used the partial derivatives

        .. math::

            \frac{\partial f_1}{\partial v_1}&= 1\\
            \frac{\partial f_2}{\partial v_2}&= \cos(v_2 t) t \\
            \frac{\partial f_1}{\partial v_2}=
            \frac{\partial f_2}{\partial v_1}&= 0

        and the fact that both :math:`H_1=Z\otimes Z` and :math:`H_2=I\otimes X`
        have two unique eigenvalues and therefore admit a two-term parameter-shift rule
        (see e.g. `Schuld et al. (2018) <https://arxiv.org/abs/1811.11184>`_).

        As a second scenario, consider the single-qubit pulse generated by

        .. math::

            H((v_1, v_2), t) = v_1 \sin(v_2 t) X

        together with the observable :math:`B=Z`.
        You may already notice that this pulse can be rewritten as a :class:`~.RX` rotation,
        because we have a single Hamiltonian term and the spectrum of :math:`H` consequently
        will be constant up to rescaling.
        In particular, the unitary time evolution under the Schrödinger equation is given by

        .. math::

            U_{(v_1, v_2)}(t_2, t_1) &=
            \exp\left(-i\int_{t_1}^{t_2} \mathrm{d}\tau v_1 \sin(v_2 \tau) X\right)\\
            &=\exp(-i\theta(v_1, v_2) X)\\
            \theta(v_1, v_2) &= \int_{t_1}^{t_2} \mathrm{d}\tau v_1 \sin(v_2 \tau)\\
            &=-\frac{v_1}{v_2}(\cos(v_2 t_2) - \cos(v_2 t_1)).

        As the ``RX`` rotation satisfies a (non-stochastic) two-term parameter-shift rule,
        we could compute the derivatives with respect to :math:`v_1` and :math:`v_2` by
        implementing :math:`\exp(-i\theta(v_1, v_2) X)`, applying the two-term shift rule
        and evaluating the classical Jacobian of the mapping :math:`\theta(v_1, v_2)`.

        Using the stochastic parameter-shift rule instead will lead to approximation errors.
        This is because the approximated integral not only includes the shifted circuit
        evaluations, which do not depend on :math:`\tau` in this example, but also on the
        classical Jacobian, which is *not* constant over :math:`\tau`.
        Therefore, it is important to implement pulses in the simplest way possible.
    """
    # pylint:disable=unused-argument
    if not has_jax:  # pragma: no cover
        raise ImportError(
            "Module jax is required for the stochastic pulse parameter-shift gradient transform. "
            "You can install jax via: pip install jax jaxlib"
        )
    if not qml.active_return():
        raise NotImplementedError(
            "The stochastic pulse parameter-shift gradient only supports the new "
            "return type. Use qml.enable_return() to turn it on."
        )
    if any(isinstance(m, VarianceMP) for m in tape.measurements):
        raise ValueError(
            "Computing the gradient of variances with the stochastic pulse "
            "parameter-shift gradient is not implemented."
        )
    if any(isinstance(m, (StateMP, VnEntropyMP, MutualInfoMP)) for m in tape.measurements):
        raise ValueError(
            "Computing the gradient of circuits that return the state is not supported."
        )

    if num_split_times < 1:
        raise ValueError(
            "Expected a positive number of samples for the stochastic pulse "
            f"parameter-shift gradient, got {num_split_times}."
        )

    if argnum is None and not tape.trainable_params:
        return _no_trainable_grad_new(tape, shots)

    if use_broadcasting and tape.batch_size is not None:
        raise ValueError("Broadcasting is not supported for tapes that already are broadcasted.")

    gradient_analysis(tape, grad_fn=stoch_pulse_grad)
    method = "analytic"
    diff_methods = grad_method_validation(method, tape)

    if all(g == "0" for g in diff_methods):
        return _all_zero_grad_new(tape, shots)

    method_map = choose_grad_methods(diff_methods, argnum)

    argnum = [i for i, dm in method_map.items() if dm == "A"]

    sampler_seed = sampler_seed or np.random.randint(18421)
    key = jax.random.PRNGKey(sampler_seed)

    return _expval_stoch_pulse_grad(tape, argnum, num_split_times, key, shots, use_broadcasting)


def _generate_tapes_and_cjacs(tape, idx, key, num_split_times, use_broadcasting):
    """Generate the tapes and compute the classical Jacobians for one given
    generating Hamiltonian term of one pulse.
    """
    op, op_idx, term_idx = tape.get_operation(idx)
    if not isinstance(op, ParametrizedEvolution):
        raise ValueError(
            "stoch_pulse_grad does not support differentiating parameters of "
            "other operations than pulses."
        )

    coeff, ob = op.H.coeffs_parametrized[term_idx], op.H.ops_parametrized[term_idx]
    if not qml.pauli.is_pauli_word(ob):
        raise ValueError(
            "stoch_pulse_grad currently only supports Pauli words as parametrized "
            f"terms in ParametrizedHamiltonian. Got {ob}"
        )
    word = qml.pauli.pauli_word_to_string(ob)
    cjac_fn = jax.jacobian(coeff, argnums=0)

    t0, *_, t1 = op.t
    taus = jnp.sort(jax.random.uniform(key, shape=(num_split_times,)) * (t1 - t0) + t0)
    cjacs = [cjac_fn(op.data[term_idx], tau) for tau in taus]
    if use_broadcasting:
        split_evolve_ops = _split_evol_ops(op, word, ob.wires, taus)
        tapes = _split_evol_tapes(tape, split_evolve_ops, op_idx)
    else:
        tapes = []
        for tau in taus:
            split_evolve_ops = _split_evol_ops(op, word, ob.wires, tau)
            tapes.extend(_split_evol_tapes(tape, split_evolve_ops, op_idx))
    avg_prefactor = (t1 - t0) / num_split_times
    return cjacs, tapes, avg_prefactor


# pylint: disable=too-many-arguments
def _expval_stoch_pulse_grad(tape, argnum, num_split_times, key, shots, use_broadcasting):
    r"""Compute the gradient of a quantum circuit composed of pulse sequences that measures
    an expectation value or probabilities, by applying the stochastic parameter shift rule.
    See the main function for the signature.
    """
    tapes = []
    gradient_data = []
    for idx, trainable_idx in enumerate(tape.trainable_params):
        if trainable_idx not in argnum:
            # Only the number of tapes is needed to indicate a zero gradient entry
            gradient_data.append((0, None, None))
            continue

        key, _key = jax.random.split(key)
        cjacs, _tapes, avg_prefactor = _generate_tapes_and_cjacs(
            tape, idx, _key, num_split_times, use_broadcasting
        )

        gradient_data.append((len(_tapes), qml.math.stack(cjacs), avg_prefactor))
        tapes.extend(_tapes)

    def processing_fn(results):
        num_measurements = len(tape.measurements)
        single_measure = num_measurements == 1
        shot_vector = isinstance(shots, Sequence)
        start = 0
        grads = []
        for num_tapes, cjacs, avg_prefactor in gradient_data:
            if num_tapes == 0:
                grads.append(None)
                continue
            res = results[start : start + num_tapes]
            start += num_tapes
            # Apply the postprocessing of the parameter-shift rule and contract
            # with classical Jacobian, effectively computing the integral approximation
            g = _parshift_and_integrate(
                res, cjacs, avg_prefactor, single_measure, shot_vector, use_broadcasting
            )
            grads.append(g)

        # g will have been defined at least once (because otherwise all gradients would have
        # been zero), providing a representative for a zero gradient to emulate its type/shape.
        zero_rep = _make_zero_rep(g, single_measure, shot_vector)

        # Fill in zero-valued gradients
        grads = [zero_rep if g is None else g for g in grads]

        num_params = len(tape.trainable_params)
        if single_measure and num_params == 1:
            return grads[0]
        len_shot_vec = _get_num_copies(shots) if shot_vector else None
        if single_measure and shot_vector:
            return _reorder_grad_axes_single_measure_shot_vector(grads, num_params, len_shot_vec)
        if not single_measure:
            return _reorder_grad_axes_multi_measure(
                grads,
                num_params,
                num_measurements,
                len_shot_vec,
                shot_vector,
            )
        return tuple(grads)

    return tapes, processing_fn


def expand_invalid_trainable_stoch_pulse_grad(x, *args, **kwargs):
    r"""Do not expand any operation. We expect the ``stoch_pulse_grad`` to be used
    on pulse programs and we do not expect decomposition pipelines between pulses
    and gate-based circuits yet.
    """
    # pylint:disable=unused-argument
    return x


stoch_pulse_grad = gradient_transform(
    _stoch_pulse_grad, expand_fn=expand_invalid_trainable_stoch_pulse_grad
)



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the :class:`ExecutionConfig` data class.
"""
from dataclasses import dataclass
from typing import Optional, Tuple, Union

from pennylane.interfaces import SUPPORTED_INTERFACES
from pennylane.gradients import SUPPORTED_GRADIENT_KWARGS

SUPPORTED_GRADIENT_METHODS = [
    "best",
    "parameter-shift",
    "backprop",
    "finite-diff",
    "device",
    "adjoint",
]


@dataclass
class ExecutionConfig:
    """
    A class to configure the execution of a quantum circuit on a device.

    See the Attributes section to learn more about the various configurable options.
    """

    shots: Optional[Union[int, Tuple[int]]] = None
    """The number of shots for an execution"""

    gradient_method: Optional[str] = None
    """The method used to compute the gradient of the quantum circuit being executed"""

    gradient_keyword_arguments: dict = None
    """Arguments used to control a gradient transform"""

    device_options: dict = None
    """Various options for the device executing a quantum circuit"""

    interface: str = "autograd"
    """The machine learning framework to use"""

    derivative_order: int = 1
    """The derivative order to compute while evaluating a gradient"""

    def __post_init__(self):
        """
        Validate the configured execution options.

        Note that this hook is automatically called after init via the dataclass integration.
        """
        if self.interface not in SUPPORTED_INTERFACES:
            raise ValueError(
                f"interface must be in {SUPPORTED_INTERFACES}, got {self.interface} instead."
            )

        if (
            self.gradient_method is not None
            and self.gradient_method not in SUPPORTED_GRADIENT_METHODS
        ):
            raise ValueError(
                f"gradient_method must be in {SUPPORTED_GRADIENT_METHODS}, got {self.gradient_method} instead."
            )

        if self.device_options is None:
            self.device_options = {}

        if self.gradient_keyword_arguments is None:
            self.gradient_keyword_arguments = {}

        if any(arg not in SUPPORTED_GRADIENT_KWARGS for arg in self.gradient_keyword_arguments):
            raise ValueError(
                f"All gradient_keyword_arguments keys must be in {SUPPORTED_GRADIENT_KWARGS}, got unexpected values: {set(self.gradient_keyword_arguments) - set(SUPPORTED_GRADIENT_KWARGS)}"
            )


DefaultExecutionConfig = ExecutionConfig()



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file contains the definition of the dot function, which computes the dot product between
a vector and a list of operators.
"""
from collections import defaultdict
from typing import Sequence, Union, Callable

import pennylane as qml
from pennylane.operation import Operator, Tensor
from pennylane.pulse import ParametrizedHamiltonian


def dot(
    coeffs: Sequence[Union[float, Callable]], ops: Sequence[Operator], pauli=False
) -> Union[Operator, ParametrizedHamiltonian]:
    r"""Returns the dot product between the ``coeffs`` vector and the ``ops`` list of operators.

    This function returns the following linear combination: :math:`\sum_{k} c_k O_k`, where
    :math:`c_k` and :math:`O_k` are the elements inside the ``coeffs`` and ``ops`` arguments, respectively.

    Args:
        coeffs (Sequence[float, Callable]): sequence containing the coefficients of the linear combination
        ops (Sequence[Operator]): sequence containing the operators of the linear combination
        pauli (bool, optional): If ``True``, a :class:`~.PauliSentence`
            operator is used to represent the linear combination. If False, a :class:`Sum` operator
            is returned. Defaults to ``False``.

    Raises:
        ValueError: if the number of coefficients and operators does not match or if they are empty

    Returns:
        Operator or ParametrizedHamiltonian: operator describing the linear combination

    **Example**

    >>> coeffs = np.array([1.1, 2.2])
    >>> ops = [qml.PauliX(0), qml.PauliY(0)]
    >>> qml.dot(coeffs, ops)
    (1.1*(PauliX(wires=[0]))) + (2.2*(PauliY(wires=[0])))
    >>> qml.dot(coeffs, ops, pauli=True)
    1.1 * X(0)
    + 2.2 * Y(0)

    ``pauli=True`` can be used to construct a more efficient, simplified version of the operator.
    Note that it returns a :class:`~.PauliSentence`, which is not an :class:`~.Operator`. This
    specialized representation can be converted to an operator:

    >>> qml.dot([1, 2], [qml.PauliX(0), qml.PauliX(0)], pauli=True).operation()
    3.0*(PauliX(wires=[0]))

    Using ``pauli=True`` and then converting the result to an :class:`~.Operator` is much faster
    than using ``pauli=False``, but it only works for pauli words
    (see :func:`~.is_pauli_word`).

    If any of the parameters listed in ``coeffs`` are callables, the resulting dot product will be a
    :class:`~.ParametrizedHamiltonian`:

    >>> coeffs = [lambda p, t: p * jnp.sin(t) for _ in range(2)]
    >>> ops = [qml.PauliX(0), qml.PauliY(0)]
    >>> qml.dot(coeffs, ops)
    ParametrizedHamiltonian: terms=2
    """
    if isinstance(coeffs, qml.QNodeCollection) or isinstance(ops, qml.QNodeCollection):
        return qml.collections.dot(coeffs, ops)

    if len(coeffs) != len(ops):
        raise ValueError("Number of coefficients and operators does not match.")
    if len(coeffs) == 0 and len(ops) == 0:
        raise ValueError("Cannot compute the dot product of an empty sequence.")

    if any(callable(c) for c in coeffs):
        return ParametrizedHamiltonian(coeffs, ops)

    if pauli:
        return _pauli_dot(coeffs, ops)

    # When casting a Hamiltonian to a Sum, we also cast its inner Tensors to Prods
    ops = [qml.prod(*op.obs) if isinstance(op, Tensor) else op for op in ops]

    if coeffs[0] != 1 and qml.math.allequal(coeffs[0], coeffs):
        # Coefficients have the same value (different to 1)
        return qml.s_prod(coeffs[0], ops[0] if len(ops) == 1 else qml.sum(*ops))

    abs_coeffs = qml.math.abs(coeffs)
    if abs_coeffs[0] != 1 and qml.math.allequal(abs_coeffs[0], abs_coeffs):
        # Coefficients have the same absolute value (different to 1)
        gcd = abs(coeffs[0])
        coeffs = [c / gcd for c in coeffs]
        return qml.s_prod(gcd, qml.dot(coeffs, ops))

    operands = [op if coeff == 1 else qml.s_prod(coeff, op) for coeff, op in zip(coeffs, ops)]
    return operands[0] if len(operands) == 1 else qml.sum(*operands)


def _pauli_dot(coeffs: Sequence[float], ops: Sequence[Operator]):
    pauli_words = defaultdict(lambda: 0)
    for coeff, op in zip(coeffs, ops):
        sentence = qml.pauli.pauli_sentence(op)
        for pw in sentence:
            pauli_words[pw] += sentence[pw] * coeff

    return qml.pauli.PauliSentence(pauli_words)



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-few-public-methods,function-redefined

"""
This file contains the ``ParametrizedEvolution`` operator.
"""

from typing import List, Union
import warnings

import pennylane as qml
from pennylane.operation import AnyWires, Operation

from .parametrized_hamiltonian import ParametrizedHamiltonian

has_jax = True
try:
    import jax
    import jax.numpy as jnp
    from jax.experimental.ode import odeint

    from .parametrized_hamiltonian_pytree import ParametrizedHamiltonianPytree
except ImportError as e:
    has_jax = False


class ParametrizedEvolution(Operation):
    r"""
    ParametrizedEvolution(H, params=None, t=None, return_intermediate=False, complementary=False, do_queue=True, id=None, **odeint_kwargs)

    Parametrized evolution gate, created by passing a :class:`~.ParametrizedHamiltonian` to
    the :func:`~.pennylane.evolve` function

    For a time-dependent Hamiltonian of the form

    .. math:: H(\{v_j\}, t) = H_\text{drift} + \sum_j f_j(v_j, t) H_j

    it implements the corresponding time-evolution operator :math:`U(t_0, t_1)`, which is the
    solution to the time-dependent Schrodinger equation.

    .. math:: \frac{d}{dt}U(t) = -i H(\{v_j\}, t) U(t).

    Under the hood, it is using a numerical ordinary differential equation (ODE) solver. It requires ``jax``,
    and will not work with other machine learning frameworks typically encountered in PennyLane.

    Args:
        H (ParametrizedHamiltonian): Hamiltonian to evolve
        params (Optional[list]): trainable parameters, passed as list where each element corresponds to
            the parameters of a scalar-valued function of the Hamiltonian being evolved.
        t (Union[float, List[float]]): If a float, it corresponds to the duration of the evolution.
            If a list of floats, the ODE solver will use all the provided time values, and
            perform intermediate steps if necessary. It is recommended to just provide a start
            and end time unless matrices of the time evolution at intermediate times need
            to be computed. Note that such absolute times only have meaning within an instance of
            ``ParametrizedEvolution`` and will not affect other gates.
            To return the matrix at intermediate evolution times, activate ``return_intermediate``
            (see below).
        do_queue (bool): determines if the scalar product operator will be queued. Default is True.
        id (str or None): id for the scalar product operator. Default is None.

    Keyword Args:
        atol (float, optional): Absolute error tolerance for the ODE solver. Defaults to ``1.4e-8``.
        rtol (float, optional): Relative error tolerance for the ODE solver. The error is estimated
            from comparing a 4th and 5th order Runge-Kutta step in the Dopri5 algorithm. This error
            is guaranteed to stay below ``tol = atol + rtol * abs(y)`` through adaptive step size
            selection. Defaults to 1.4e-8.
        mxstep (int, optional): maximum number of steps to take for each timepoint for the
            ODE solver. Defaults to ``jnp.inf``.
        hmax (float, optional): maximum step size allowed for the ODE solver. Defaults to ``jnp.inf``.
        return_intermediate (bool): Whether or not the ``matrix`` method returns all intermediate
            solutions of the time evolution at the times provided in ``t = [t_0,...,t_f]``.
            If ``False`` (the default), only the matrix for the full time evolution is returned.
            If ``True``, all solutions including the initial condition are returned;
            when used in a circuit, this results in ``ParametrizedEvolution`` being a broadcasted
            operation, see the usage details ("Computing intermediate time evolution") below.
        complementary (bool): Whether or not to compute the complementary time evolution when using
            ``return_intermediate=True`` (ignored otherwise).
            If ``False`` (the default), the usual solutions to the Schrodinger equation
            :math:`\{U(t_0, t_0), U(t_0, t_1),\dots, U(t_0, t_f)\}` are computed,
            where :math:`t_i` are the additional times provided in ``t``.
            If ``True``, the *remaining* time evolution to :math:`t_f` is computed instead, returning
            :math:`\{U(t_0, t_f), U(t_1, t_f),\dots, U(t_{f-1}, t_f), U(t_f, t_f)\}`.

    .. warning::
        The :class:`~.ParametrizedHamiltonian` must be Hermitian at all times. This is not explicitly checked
        when creating a :class:`~.ParametrizedEvolution` from the :class:`~.ParametrizedHamiltonian`.

    **Example**

    To create a :class:`~.ParametrizedEvolution`, we first define a :class:`~.ParametrizedHamiltonian`
    describing the system, and then pass it to :func:`~pennylane.evolve`:

    .. code-block:: python

        from jax import numpy as jnp

        f1 = lambda p, t: jnp.sin(p * t)
        H = f1 * qml.PauliY(0)

        ev = qml.evolve(H)

    The initial :class:`~.ParametrizedEvolution` does not have set parameters, and so will not
    have a matrix defined. To obtain an Operator with a matrix, it must be passed parameters and
    a time interval:

    >>> qml.matrix(ev([1.2], t=[0, 4]))
    Array([[ 0.72454906+0.j, -0.6892243 +0.j],
           [ 0.6892243 +0.j,  0.72454906+0.j]], dtype=complex64)

    The parameters can be updated by calling the :class:`~.ParametrizedEvolution` again with different inputs.

    When calling the :class:`~.ParametrizedEvolution`, keyword arguments can be passed to specify
    behaviour of the ODE solver.

    The :class:`~.ParametrizedEvolution` can be implemented in a QNode:

    .. code-block:: python

        import jax

        dev = qml.device("default.qubit.jax", wires=1)
        @jax.jit
        @qml.qnode(dev, interface="jax")
        def circuit(params):
            qml.evolve(H)(params, t=[0, 10])
            return qml.expval(qml.PauliZ(0))

    >>> params = [1.2]
    >>> circuit(params)
    Array(0.96632576, dtype=float32)

    >>> jax.grad(circuit)(params)
    [Array(2.3569832, dtype=float32)]

    .. note::
        In the example above, the decorator ``@jax.jit`` is used to compile this execution just-in-time. This means
        the first execution will typically take a little longer with the benefit that all following executions
        will be significantly faster, see the jax docs on jitting. JIT-compiling is optional, and one can remove
        the decorator when only single executions are of interest.

    .. warning::

        The time argument ``t`` corresponds to the time window used to compute the scalar-valued
        functions present in the :class:`ParametrizedHamiltonian` class. Consequently, executing
        two ``ParametrizedEvolution`` operators using the same time window does not mean both
        operators are executed simultaneously, but rather that both evaluate their respective
        scalar-valued functions using the same time window. See Usage Details.

    .. note::

        Using ``return_intermediate`` in a quantum circuit leads to broadcasted execution,
        which can lead to unintended additional computational cost.
        Also consider the usage details below.

    .. details::
        :title: Usage Details

        The parameters used when calling the ``ParametrizedEvolution`` are expected to have the same order
        as the functions used to define the :class:`~.ParametrizedHamiltonian`. For example:

        .. code-block:: python3

            def f1(p, t):
                return jnp.sin(p[0] * t**2) + p[1]

            def f2(p, t):
                return p * jnp.cos(t)

            H = 2 * qml.PauliX(0) + f1 * qml.PauliY(0) + f2 * qml.PauliZ(0)
            ev = qml.evolve(H)

        >>> params = [[4.6, 2.3], 1.2]
        >>> qml.matrix(ev(params, t=0.5))
        Array([[-0.18354285-0.26303384j, -0.7271658 -0.606923j  ],
               [ 0.7271658 -0.606923j  , -0.18354285+0.26303384j]],      dtype=complex64)

        Internally the solver is using ``f1([4.6, 2.3], t)`` and ``f2(1.2, t)`` at each timestep when
        finding the matrix.

        In the case where we have defined two Hamiltonians, ``H1`` and ``H2``, and we want to find a time evolution
        where the two are driven simultaneously for some period of time, it is important that both are included in
        the same call of :func:`~.pennylane.evolve`.
        For non-commuting operations, applying ``qml.evolve(H1)(params, t=[0, 10])`` followed by
        ``qml.evolve(H2)(params, t=[0, 10])`` will **not** apply the two pulses simultaneously, despite the overlapping
        time window. Instead, it will execute ``H1`` in the ``[0, 10]`` time window, and then subsequently execute
        ``H2`` using the same time window to calculate the evolution, but without taking into account how the time
        evolution of ``H1`` affects the evolution of ``H2`` and vice versa.

        Consider two non-commuting :class:`ParametrizedHamiltonian` objects:

        .. code-block:: python

            from jax import numpy as jnp

            ops = [qml.PauliX(0), qml.PauliY(1), qml.PauliZ(2)]
            coeffs = [lambda p, t: p for _ in range(3)]
            H1 = qml.dot(coeffs, ops)  # time-independent parametrized Hamiltonian

            ops = [qml.PauliZ(0), qml.PauliY(1), qml.PauliX(2)]
            coeffs = [lambda p, t: p * jnp.sin(t) for _ in range(3)]
            H2 = qml.dot(coeffs, ops) # time-dependent parametrized Hamiltonian

        The evolutions of the :class:`ParametrizedHamiltonian` can be used in a QNode.

        .. code-block:: python

            dev = qml.device("default.qubit.jax", wires=3)

            @qml.qnode(dev, interface="jax")
            def circuit1(params):
                qml.evolve(H1)(params, t=[0, 10])
                qml.evolve(H2)(params, t=[0, 10])
                return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1) @ qml.PauliZ(2))

            @qml.qnode(dev, interface="jax")
            def circuit2(params):
                qml.evolve(H1 + H2)(params, t=[0, 10])
                return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1) @ qml.PauliZ(2))

        In ``circuit1``, the two Hamiltonians are evolved over the same time window, but inside different operators.
        In ``circuit2``, we add the two to form a single :class:`~.ParametrizedHamiltonian`. This will combine the
        two so that the expected parameters will be ``params1 + params2`` (as an addition of ``list``).
        They can then be included inside a single :class:`~.ParametrizedEvolution`.

        The resulting evolutions of ``circuit1`` and ``circuit2`` are **not** identical:

        >>> params = jnp.array([1., 2., 3.])
        >>> circuit1(params)
        Array(-0.01543971, dtype=float32)

        >>> params = jnp.concatenate([params, params])  # H1 + H2 requires 6 parameters!
        >>> circuit2(params)
        Array(-0.78236955, dtype=float32)

        Here, ``circuit1`` is not executing the evolution of ``H1`` and ``H2`` simultaneously, but rather
        executing ``H1`` in the ``[0, 10]`` time window and then executing ``H2`` with the same time window,
        without taking into account how the time evolution of ``H1`` affects the evolution of ``H2`` and vice versa!

        One can also provide a list of time values that the ODE solver will use to calculate the evolution of the
        ``ParametrizedHamiltonian``. Keep in mind that the ODE solver uses an adaptive step size, thus
        it might use additional intermediate time values.

        .. code-block:: python

            t = jnp.arange(0., 10.1, 0.1)
            @qml.qnode(dev, interface="jax")
            def circuit(params):
                qml.evolve(H1 + H2)(params, t=t)
                return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1) @ qml.PauliZ(2))

        >>> circuit(params)
        Array(-0.78236955, dtype=float32)
        >>> jax.grad(circuit)(params)
        Array([-4.8066125 ,  3.703827  , -1.3297377 , -2.406232  ,  0.6811726 ,
            -0.52277344], dtype=float32)

        Given that we used the same time window (``[0, 10]``), the results are the same as before.

        **Computing intermediate time evolution**

        As discussed above, the ODE solver will evaluate the Schrodinger equation at
        intermediate times in any case. By passing additional time values explicitly in the time
        window ``t`` and setting ``return_intermediate=True``, the ``matrix`` method will
        return the matrices for the intermediate time evolutions as well:

        .. math::

            \{U(t_0, t_0), U(t_0, t_1), \dots, U(t_0, t_{f-1}), U(t_0, t_f)\}.

        The first entry here is the initial condition :math:`U(t_0, t_0)=1`. For a simple
        time-dependent single-qubit Hamiltonian, this feature looks like the following:

        .. code-block:: python

            ops = [qml.PauliZ(0), qml.PauliY(0), qml.PauliX(0)]
            coeffs = [lambda p, t: p * jnp.cos(t) for _ in range(3)]
            H = qml.dot(coeffs, ops) # time-dependent parametrized Hamiltonian

            param = [jnp.array(0.2), jnp.array(1.1), jnp.array(-1.3)]
            time = jnp.linspace(0.1, 0.4, 6) # Six time points from 0.1 to 0.4

            ev = qml.evolve(H)(param, time, return_intermediate=True)

        >>> ev_mats = ev.matrix()
        >>> ev_mats.shape
        (6, 2, 2)

        Note that the broadcasting axis has length ``len(time)`` and is the first axis of the
        returned tensor.
        We may use this feature within QNodes executed on a simulator, returning the
        measurements for all intermediate time steps:

        .. code-block:: python

            dev = qml.device("default.qubit.jax", wires=1)

            @qml.qnode(dev, interface="jax")
            def circuit(param, time):
                qml.evolve(H)(param, time, return_intermediate=True)
                return qml.probs(wires=[0])

        >>> circuit(param, time)
        Array([[1.        , 0.        ],
               [0.9897738 , 0.01022595],
               [0.9599043 , 0.04009585],
               [0.9123617 , 0.08763832],
               [0.84996957, 0.15003097],
               [0.7761489 , 0.22385144]], dtype=float32)


        **Computing complementary time evolution**

        When using ``return_intermediate=True``, the partial time evolutions share the *initial*
        time :math:`t_0`. For some applications, however, it may be useful to compute the
        complementary time evolutions, i.e. the partial evolutions that share the *final* time
        :math:`t_f`. This can be activated by setting ``complementary=True``, which will make
        ``ParametrizedEvolution.matrix`` return the matrices

        .. math::

            \{U(t_0, t_f), U(t_1, t_f), \dots, U(t_f, t_f)\}.

        Using the Hamiltonian from the example above:

        >>> complementary_ev = ev(param, time, return_intermediate=True, complementary=True)
        >>> comp_ev_mats = complementary_ev.matrix()
        >>> comp_ev_mats.shape
        (6, 2, 2)

        If we multiply the matrices computed before with ``complementary=False`` with these
        complementary evolution matrices from the left, we obtain the full time evolution,
        which we can check by comparing to the last entry of ``ev_mats``:

        >>> for mat, c_mat in zip(ev_mats, comp_ev_mats):
        ...     print(qml.math.allclose(c_mat @ mat, ev_mats[-1]))
        True
        True
        True
        True
        True
        True

    """

    _name = "ParametrizedEvolution"
    num_wires = AnyWires
    grad_method = "A"

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        H: ParametrizedHamiltonian,
        params: list = None,
        t: Union[float, List[float]] = None,
        return_intermediate: bool = False,
        complementary: bool = False,
        do_queue=True,
        id=None,
        **odeint_kwargs
    ):
        if not all(op.has_matrix or isinstance(op, qml.Hamiltonian) for op in H.ops):
            raise ValueError(
                "All operators inside the parametrized hamiltonian must have a matrix defined."
            )
        self._has_matrix = params is not None and t is not None
        self.H = H
        self.odeint_kwargs = odeint_kwargs
        if t is None:
            self.t = None
        else:
            self.t = qml.math.array([0.0, t] if qml.math.ndim(t) == 0 else t, dtype=float)
        if complementary and not return_intermediate:
            warnings.warn(
                "The keyword argument complementary does not have any effect if "
                "return_intermediate is set to False."
            )
        params = [] if params is None else params
        super().__init__(*params, wires=H.wires, do_queue=do_queue, id=id)
        self.hyperparameters["return_intermediate"] = return_intermediate
        self.hyperparameters["complementary"] = complementary
        self._check_time_batching()

    def __call__(self, params, t, return_intermediate=None, complementary=None, **odeint_kwargs):
        if not has_jax:
            raise ImportError(
                "Module jax is required for the ``ParametrizedEvolution`` class. "
                "You can install jax via: pip install jax"
            )
        # Need to cast all elements inside params to `jnp.arrays` to make sure they are not cast
        # to `np.arrays` inside `Operator.__init__`
        params = [jnp.array(p) for p in params]
        # Inherit return_intermediate and complementary from self if not provided.
        if return_intermediate is None:
            return_intermediate = self.hyperparameters["return_intermediate"]
        if complementary is None:
            complementary = self.hyperparameters["complementary"]
        odeint_kwargs = {**self.odeint_kwargs, **odeint_kwargs}
        if qml.QueuingManager.recording():
            qml.QueuingManager.remove(self)

        return ParametrizedEvolution(
            H=self.H,
            params=params,
            t=t,
            return_intermediate=return_intermediate,
            complementary=complementary,
            do_queue=True,
            id=self.id,
            **odeint_kwargs
        )

    def _check_time_batching(self):
        """Check whether the time argument is broadcasted/batched."""
        if not self.hyperparameters["return_intermediate"] or self.t is None:
            return
        # Subtract 1 because the identity is never returned by `matrix`. If `complementary=True`,
        # subtract and additional 1 because the full time evolution is not being returned.
        self._batch_size = self.t.shape[0]

    @property
    def hash(self):
        """int: Integer hash that uniquely represents the operator."""
        return hash(
            (
                str(self.name),
                tuple(self.wires.tolist()),
                str(self.hyperparameters.values()),
                str(self.t),
                str(self.data),
                self.H,
                str(self.odeint_kwargs.values()),
            )
        )

    # pylint: disable=arguments-renamed, invalid-overridden-method
    @property
    def has_matrix(self):
        return self._has_matrix

    # pylint: disable=import-outside-toplevel
    def matrix(self, wire_order=None):
        if not has_jax:
            raise ImportError(
                "Module jax is required for the ``ParametrizedEvolution`` class. "
                "You can install jax via: pip install jax"
            )
        if not self.has_matrix:
            raise ValueError(
                "The parameters and the time window are required to compute the matrix. "
                "You can update its values by calling the class: EV(params, t)."
            )
        y0 = jnp.eye(2 ** len(self.wires), dtype=complex)

        with jax.ensure_compile_time_eval():
            H_jax = ParametrizedHamiltonianPytree.from_hamiltonian(
                self.H, dense=len(self.wires) < 3, wire_order=self.wires
            )

        def fun(y, t):
            """dy/dt = -i H(t) y"""
            return (-1j * H_jax(self.data, t=t)) @ y

        mat = odeint(fun, y0, self.t, **self.odeint_kwargs)
        if self.hyperparameters["return_intermediate"] and self.hyperparameters["complementary"]:
            # Compute U(t_0, t_f)@U(t_0, t_i)^\dagger, where i indexes the first axis of mat
            mat = qml.math.tensordot(mat[-1], qml.math.conj(mat), axes=[[1], [-1]])
            # The previous line leaves the axis indexing the t_i as second, so we move it up
            mat = qml.math.moveaxis(mat, 1, 0)
        elif not self.hyperparameters["return_intermediate"]:
            mat = mat[-1]
        return qml.math.expand_matrix(mat, wires=self.wires, wire_order=wire_order)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module containing the ``JaxParametrizedHamiltonian`` class."""
from dataclasses import dataclass
from typing import Callable, Optional, Tuple, Union

import jax
import jax.numpy as jnp
from jax.experimental import sparse
from jax.tree_util import register_pytree_node_class

import pennylane as qml

from .parametrized_hamiltonian import ParametrizedHamiltonian
from .hardware_hamiltonian import HardwareHamiltonian


@register_pytree_node_class
@dataclass
class ParametrizedHamiltonianPytree:
    """Jax pytree class that represents a ``ParametrizedHamiltonian``."""

    mat_fixed: Optional[Union[jnp.ndarray, sparse.CSR]]
    mats_parametrized: Tuple[Union[jnp.ndarray, sparse.CSR], ...]
    coeffs_parametrized: Tuple[Callable]
    reorder_fn: Callable

    @staticmethod
    def from_hamiltonian(H: ParametrizedHamiltonian, *, dense: bool = False, wire_order=None):
        """Convert a ``ParametrizedHamiltonian`` into a jax pytree object.

        Args:
            H (ParametrizedHamiltonian): parametrized Hamiltonian to convert
            dense (bool, optional): Decide wether a dense/sparse matrix is used. Defaults to False.
            wire_order (list, optional): Wire order of the returned ``JaxParametrizedOperator``.
                Defaults to None.

        Returns:
            ParametrizedHamiltonianPytree: pytree object
        """
        make_array = jnp.array if dense else sparse.CSR.fromdense

        if len(H.ops_fixed) > 0:
            mat_fixed = make_array(qml.matrix(H.H_fixed(), wire_order=wire_order))
        else:
            mat_fixed = None

        mats_parametrized = tuple(
            make_array(qml.matrix(op, wire_order=wire_order)) for op in H.ops_parametrized
        )

        if isinstance(H, HardwareHamiltonian):
            return ParametrizedHamiltonianPytree(
                mat_fixed,
                mats_parametrized,
                H.coeffs_parametrized,
                reorder_fn=H.reorder_fn,
            )

        return ParametrizedHamiltonianPytree(
            mat_fixed, mats_parametrized, H.coeffs_parametrized, reorder_fn=None
        )

    def __call__(self, pars, t):
        if self.mat_fixed is not None:
            ops = (self.mat_fixed,)
            coeffs = (1,)
        else:
            ops = ()
            coeffs = ()

        if self.reorder_fn:
            pars = self.reorder_fn(pars, self.coeffs_parametrized)
        coeffs = coeffs + tuple(f(p, t) for f, p in zip(self.coeffs_parametrized, pars))
        return LazyDotPytree(coeffs, ops + self.mats_parametrized)

    def tree_flatten(self):
        """Function used by ``jax`` to flatten the JaxParametrizedOperator.

        Returns:
            tuple: tuple containing the matrices and the parametrized coefficients defining the class
        """
        matrices = (self.mat_fixed, self.mats_parametrized)
        param_coeffs = self.coeffs_parametrized
        reorder_fn = self.reorder_fn
        return (matrices, param_coeffs, reorder_fn)

    @classmethod
    def tree_unflatten(cls, param_coeffs: tuple, matrices: tuple, reorder_fn: callable):
        """Function used by ``jax`` to unflatten the JaxParametrizedOperator.

        Args:
            param_coeffs (tuple): tuple containing the parametrized coefficients of the class
            matrices (tuple): tuple containing the matrices of the class
            reorder_fn(callable): callable or None indicating how parameters should be
                re-orderd to pass to the __call__ method

        Returns:
            JaxParametrizedOperator: a JaxParametrizedOperator instance
        """
        return cls(*matrices, param_coeffs, reorder_fn)


@register_pytree_node_class
@dataclass
class LazyDotPytree:
    """Jax pytree representing a lazy dot operation."""

    coeffs: Tuple[complex, ...]
    mats: Tuple[Union[jnp.ndarray, sparse.CSR], ...]

    @jax.jit
    def __matmul__(self, other):
        return sum(c * (m @ other) for c, m in zip(self.coeffs, self.mats))

    def __mul__(self, other):
        if jnp.array(other).ndim == 0:
            return LazyDotPytree(tuple(other * c for c in self.coeffs), self.mats)
        return NotImplemented

    __rmul__ = __mul__

    def tree_flatten(self):
        """Function used by ``jax`` to flatten the JaxLazyDot operation.

        Returns:
            tuple: tuple containing children and the auxiliary data of the class
        """
        children = (self.coeffs, self.mats)
        aux_data = None
        return (children, aux_data)

    # pylint: disable=unused-argument
    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """Function used by ``jax`` to unflatten the ``JaxLazyDot`` pytree.

        Args:
            aux_data (None): empty argument
            children (tuple): tuple containing the coefficients and the matrices of the operation

        Returns:
            JaxLazyDot: JaxLazyDot instance
        """
        return cls(*children)



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains the classes/functions specific for simulation of superconducting transmon hardware systems"""
import warnings

from dataclasses import dataclass
from typing import Callable, List, Union

import pennylane as qml
import pennylane.numpy as np
from pennylane.pulse import HardwareHamiltonian
from pennylane.pulse.hardware_hamiltonian import HardwarePulse
from pennylane.typing import TensorLike
from pennylane.wires import Wires


# TODO ladder operators once there is qudit support
# pylint: disable=unused-argument
def a(wire, d=2):
    """creation operator"""
    return qml.s_prod(0.5, qml.PauliX(wire)) + qml.s_prod(0.5j, qml.PauliY(wire))


def ad(wire, d=2):
    """annihilation operator"""
    return qml.s_prod(0.5, qml.PauliX(wire)) + qml.s_prod(-0.5j, qml.PauliY(wire))


# pylint: disable=too-many-arguments
def transmon_interaction(
    qubit_freq: Union[float, list],
    connections: list,
    coupling: Union[float, list],
    wires: list,
    anharmonicity=None,
    d=2,
):
    r"""Returns a :class:`ParametrizedHamiltonian` representing the circuit QED Hamiltonian of a
    superconducting transmon system.

    The Hamiltonian is given by

    .. math::

        H = \sum_{q\in \text{wires}} \omega_q a^\dagger_q a_q
        + \sum_{(i, j) \in \mathcal{C}} g_{ij} \left(a^\dagger_i a_j + a_j^\dagger a_i \right)
        + \sum_{q\in \text{wires}} \alpha_q a^\dagger_q a^\dagger_q a_q a_q

    where :math:`[a^\dagger_p, a_q] = i \delta_{pq}` are bosonic creation and annihilation operators.
    The first term describes the effect of the dressed qubit frequencies ``qubit_freq`` :math:`= \omega_q/ (2\pi)`,
    the second term their ``coupling`` :math:`= g_{ij}/(2\pi)` and the last the
    ``anharmonicity`` :math:`= \alpha_q/(2\pi)`, which all can vary for
    different qubits. In practice, the bosonic operators are restricted to a finite dimension of the
    local Hilbert space (default ``d=2`` corresponds to qubits).
    In that case, the anharmonicity is set to :math:`\alpha=0` and ignored.

    The values of :math:`\omega` and :math:`\alpha` are typically around :math:`5 \times 2\pi \text{GHz}`
    and :math:`0.3 \times 2\pi \text{GHz}`, respectively.
    It is common for different qubits to be out of tune with different energy gaps. The coupling strength
    :math:`g` typically varies betwewen :math:`[0.001, 0.1] \times 2\pi \text{GHz}`. For some example parameters,
    see e.g. `arXiv:1804.04073 <https://arxiv.org/abs/1804.04073>`_,
    `arXiv:2203.06818 <https://arxiv.org/abs/2203.06818>`_, or `arXiv:2210.15812 <https://arxiv.org/abs/2210.15812>`_.

    .. note:: Currently only supporting ``d=2`` with qudit support planned in the future.

    .. seealso::

        :func:`~.transmon_drive`

    Args:
        qubit_freq (Union[float, list[float]]): List of dressed qubit frequencies. This should be in units
            of frequency (GHz), and will be converted to angular frequency :math:`\omega` internally where
            needed, i.e. multiplied by :math:`2 \pi`. When passing a single float all qubits are assumed to
            have that same frequency.
        connections (list[tuple(int)]): List of connections ``(i, j)`` between qubits i and j.
            When the wires in ``connections`` are not contained in ``wires``, a warning is raised.
        coupling (Union[float, list[float]]): List of coupling strengths. This should be in units
            of frequency (GHz), and will be converted to angular frequency internally where
            needed, i.e. multiplied by :math:`2 \pi`. Needs to match the length of ``connections``.
            When passing a single float need explicit ``wires``.
        anharmonicity (Union[float, list[float]]): List of anharmonicities. This should be in units
            of frequency (GHz), and will be converted to angular frequency internally where
            needed, i.e. multiplied by :math:`2 \pi`. Ignored when ``d=2``.
            When passing a single float all qubits are assumed to have that same anharmonicity.
        wires (list): Needs to be of the same length as qubit_freq. Note that there can be additional
            wires in the resulting operator from the ``connections``, which are treated independently.
        d (int): Local Hilbert space dimension. Defaults to ``d=2`` and is currently the only supported value.

    Returns:
        :class:`~.ParametrizedHamiltonian`: a :class:`~.ParametrizedHamiltonian` representing the transmon interaction

    **Example**

    We can set up the transmon interaction Hamiltonian with uniform coefficients by passing ``float`` values.

    .. code-block::

        connections = [[0, 1], [1, 3], [2, 1], [4, 5]]
        H = qml.pulse.transmon_interaction(qubit_freq=0.5, connections=connections, coupling=1., wires=range(6))

    The resulting :class:`~.ParametrizedHamiltonian` consists of ``4`` coupling terms and ``6`` qubits
    because there are six different wire indices in ``connections``.

    >>> print(H)
    ParametrizedHamiltonian: terms=10

    We can also provide individual values for each of the qubit energies and coupling strengths, here of order :math:`0.1 \times 2\pi\text{GHz}` and :math:`1 \times 2\pi\text{GHz}`, respectively.

    .. code-block::

        qubit_freqs = [0.5, 0.4, 0.3, 0.2, 0.1, 0.]
        couplings= [1., 2., 3., 4.]
        H = qml.pulse.transmon_interaction(qubit_freq=qubit_freqs,
                                           connections=connections,
                                           coupling=couplings,
                                           wires=range(6))

    The interaction term is dependent only on the typically fixed transmon energies and coupling strengths.
    Executing this as a pulse program via :func:`~.evolve` would correspond to all driving fields being turned off.
    To add a driving field, see :func:`~.transmon_drive`.

    """
    if d != 2:
        raise NotImplementedError(
            "Currently only supporting qubits. Qutrits and qudits are planned in the future."
        )

    # if wires is None and qml.math.ndim(omega) == 0:
    #     raise ValueError(
    #         f"Cannot instantiate wires automatically. Either need specific wires or a list of omega."
    #         f"Received wires {wires} and omega of type {type(omega)}"
    #     )

    # wires = wires or list(range(len(omega)))

    n_wires = len(wires)

    if not Wires(wires).contains_wires(Wires(np.unique(connections).tolist())):
        warnings.warn(
            f"Caution, wires and connections do not match. "
            f"I.e., wires in connections {connections} are not contained in the wires {wires}"
        )

    # Prepare coefficients
    if anharmonicity is None:
        anharmonicity = [0.0] * n_wires

    # TODO: make coefficients callable / trainable. Currently not supported
    if qml.math.ndim(qubit_freq) == 0:
        qubit_freq = [qubit_freq] * n_wires
    if len(qubit_freq) != n_wires:
        raise ValueError(
            f"Number of qubit frequencies in {qubit_freq} does not match the provided wires = {wires}"
        )

    if qml.math.ndim(coupling) == 0:
        coupling = [coupling] * len(connections)
    if len(coupling) != len(connections):
        raise ValueError(
            f"Number of coupling terms {coupling} does not match the provided connections = {connections}"
        )

    settings = TransmonSettings(connections, qubit_freq, coupling, anharmonicity=anharmonicity)

    omega = [callable_freq_to_angular(f) if callable(f) else (2 * np.pi * f) for f in qubit_freq]
    g = [callable_freq_to_angular(c) if callable(c) else (2 * np.pi * c) for c in coupling]

    # qubit term
    coeffs = list(omega)
    observables = [ad(i, d) @ a(i, d) for i in wires]

    # coupling term
    coeffs += list(g)
    observables += [ad(i, d) @ a(j, d) + ad(j, d) @ a(i, d) for (i, j) in connections]

    # TODO Qudit support. Currently not supported but will be in the future.
    # if d>2:
    #     if qml.math.ndim(anharmonicity)==0:
    #         anharmonicity = [anharmonicity] * n_wires
    #     if len(anharmonicity) != n_wires:
    #         raise ValueError(f"Number of qubit anharmonicities anharmonicity = {anharmonicity} does not match the provided wires = {wires}")
    #     # anharmonicity term
    #     alpha = [2 * np.pi * a for a in anharmonicity]
    #     coeffs += list(alpha)
    #     observables += [ad(i, d) @ ad(i, d) @ a(i, d) @ a(i, d) for i in wires]

    return HardwareHamiltonian(
        coeffs, observables, settings=settings, reorder_fn=_reorder_AmpPhaseFreq
    )


def callable_freq_to_angular(fn):
    """Add a factor of 2pi to a callable result to convert from Hz to rad/s"""

    def angular_fn(p, t):
        return 2 * np.pi * fn(p, t)

    return angular_fn


@dataclass
class TransmonSettings:
    """Dataclass that contains the information of a Transmon setup.

    .. see-also:: :func:`transmon_interaction`

    Args:
            connections (List): List `[[idx_q0, idx_q1], ..]` of connected qubits (wires)
            qubit_freq (List[float, Callable]):
            coupling (List[list, TensorLike, Callable]):
            anharmonicity (List[float, Callable]):

    """

    connections: List
    qubit_freq: Union[float, Callable]
    coupling: Union[list, TensorLike, Callable]
    anharmonicity: Union[float, Callable]

    def __eq__(self, other):
        return (
            qml.math.all(self.connections == other.connections)
            and qml.math.all(self.qubit_freq == other.qubit_freq)
            and qml.math.all(self.coupling == other.coupling)
            and qml.math.all(self.anharmonicity == other.anharmonicity)
        )

    def __add__(self, other):
        if other is not None:
            new_connections = list(self.connections) + list(other.connections)
            new_qubit_freq = list(self.qubit_freq) + list(other.qubit_freq)
            new_coupling = list(self.coupling) + list(other.coupling)
            new_anh = list(self.anharmonicity) + list(other.anharmonicity)
            return TransmonSettings(
                new_connections, new_qubit_freq, new_coupling, anharmonicity=new_anh
            )

        return self


def transmon_drive(amplitude, phase, freq, wires, d=2):
    r"""Returns a :class:`ParametrizedHamiltonian` representing the drive term of a transmon qubit.

    The Hamiltonian is given by

    .. math::

        \Omega(t) \left(e^{i (\phi(t) + \nu t)} a_q + e^{-i (\phi(t) + \nu t)} a^\dagger_q\right)

    where :math:`[a^\dagger_p, a_q] = i \delta_{pq}` are bosonic creation and annihilation operators
    and :math:`q` is the qubit label (``wires``).
    The arguments ``amplitude``, ``phase`` and ``freq`` correspond to :math:`\Omega / (2\pi)`, :math:`\phi`
    and :math:`\nu / (2\pi)`, respectively, and can all be either fixed numbers (``float``) or depend on time
    (``callable``). If they are time-dependent, they need to abide by the restrictions imposed
    in :class:`ParametrizedHamiltonian` and have a signature of two parameters, ``(params, t)``.

    For realistic simulations, one may restrict the amplitude, phase and drive frequency parameters.
    For example, the authors in `2008.04302 <https://arxiv.org/abs/2008.04302>`_ impose the restrictions of
    a maximum amplitude :math:`\Omega_{\text{max}} = 20 \text{MHz}` and the carrier frequency to deviate at most
    :math:`\nu - \omega = \pm 1 \text{GHz}` from the qubit frequency :math:`\omega`
    (see :func:`~.transmon_interaction`).
    The phase :math:`\phi(t)` is typically a slowly changing function of time compared to :math:`\Omega(t)`.

    .. note:: Currently only supports ``d=2`` with qudit support planned in the future. For ``d=2`` we have :math:`a:=\frac{1}{2}(\sigma^x + i \sigma^y)`.

    .. note:: Due to convention in the respective fields, we omit the factor :math:`\frac{1}{2}` present in the related constructor :func:`~.rydberg_drive`

    .. seealso::

        :func:`~.rydberg_drive`, :func:`~.transmon_interaction`

    Args:
        amplitude (Union[float, callable]): Float or callable representing the amplitude of the driving field.
            This should be in units of frequency (GHz), and will be converted to angular frequency
            :math:`\Omega(t)` internally where needed, i.e. multiplied by :math:`2 \pi`.
        phase (Union[float, callable]): Float or callable returning phase :math:`\phi(t)` (in radians).
            Can be a fixed number (``float``) or depend on time (``callable``)
        freq (Union[float, callable]): Float or callable representing the frequency of the driving field.
            This should be in units of frequency (GHz), and will be converted to angular frequency
            :math:`\nu(t)` internally where needed, i.e. multiplied by :math:`2 \pi`.
        wires (Union[int, list[int]]): Label of the qubit that the drive acts upon. Can be a list of multiple wires.
        d (int): Local Hilbert space dimension. Defaults to ``d=2`` and is currently the only supported value.

    Returns:
        :class:`~.ParametrizedHamiltonian`: a :class:`~.ParametrizedHamiltonian` representing the transmon interaction

    **Example**

    We can construct a drive term acting on qubit ``0`` in the following way. We parametrize the amplitude and phase
    via :math:`\Omega(t)/(2 \pi) = A \times \sin^2(\pi t)` and :math:`\phi(t) = \phi_0 (t - \frac{1}{2})`. The squared
    sine ensures that the amplitude will be strictly positive (a requirement for some hardware).

    .. code-block:: python3

        def amp(A, t): return A * jnp.sin(jnp.pi*t) ** 2
        def phase(phi0, t): return phi0 * (t - 0.5)

        H = qml.pulse.transmon_drive(amp, phase, 0, 0)

        t = 0.5
        A = 0.1
        phi = 0.001
        params = [A, phi]

    Evaluated at :math:`t = \frac{1}{2}` with the parameters :math:`A = 0.1` and :math:`\phi = 10^{-3}` we obtain
    :math:`2 \pi A \left(\frac{1}{2}(\sigma^x + i \sigma^y) + \frac{1}{2}(\sigma^x + i \sigma^y)\right) = 2 \pi A \sigma^x = 0.63 \sigma^x`.

    >>> H(params, t)
    (0.6283185307179586*(PauliX(wires=[0]))) + (0.0*(-1*(PauliY(wires=[0]))))

    We can combine ``transmon_drive()`` with :func:`~.transmon_interaction` to create a full driven transmon Hamiltonian.
    Let us look at a chain of three transmon qubits that are coupled with their direct neighbors. We provide all
    frequencies in GHz (conversion to angular frequency, i.e. multiplication by :math:`2 \pi`, is taken care of
    internally where needed).

    We use values around :math:`\omega = 5 \times 2\pi \text{GHz}` for resonant frequencies, and coupling strenghts
    on the order of around :math:`g = 0.01 \times 2\pi \text{GHz}`.

    We parametrize the amplitude as a squared sinusodial and make the maximum amplitude as well as the drive frequency
    trainable parameters. We simulate the evolution for a time window of :math:`[0, 5]\text{ns}`.

    .. code-block:: python3

        qubit_freqs = [5.1, 5., 5.3]
        connections = [[0, 1], [1, 2]]  # qubits 0 and 1 are coupled, as are 1 and 2
        g = [0.02, 0.05]
        H = qml.pulse.transmon_interaction(qubit_freqs, connections, g, wires=range(3))

        def amp(max_amp, t): return max_amp * jnp.sin(t) ** 2
        def freq(fr, t): return fr
        phase = 0.
        t=2

        H += qml.pulse.transmon_drive(amp, phase, freq, 0)
        H += qml.pulse.transmon_drive(amp, phase, freq, 1)
        H += qml.pulse.transmon_drive(amp, phase, freq, 2)

        dev = qml.device("default.qubit.jax", wires=range(3))

        @jax.jit
        @qml.qnode(dev, interface="jax")
        def qnode(params):
            qml.evolve(H)(params, t=5.)
            return qml.expval(qml.PauliZ(0) + qml.PauliZ(1) + qml.PauliZ(2))

    We evaluate the Hamiltonian with some arbitrarily chosen maximum amplitudes (here on the order of :math:`0.5 \times 2\pi \text{GHz}`)
    and set the drive frequency equal to the qubit frequencies. Note how the order of the construction
    of ``H`` determines the order with which the parameters need to be passed to
    :class:`~.ParametrizedHamiltonian` and :func:`~.evolve`. We made the drive frequencies
    trainable parameters by providing a constant callable above instead of fixed values.
    This allows us to differentiate with respect to the frequencies and to optimize them.

    >>> max_amp0, max_amp1, max_amp2 = [0.5, 0.3, 0.6]
    >>> fr0, fr1, fr2 = qubit_freqs
    >>> params = [max_amp0, fr0, max_amp1, fr1, max_amp2, fr2]
    >>> qnode(params)
    Array(-1.44726531, dtype=float64)
    >>> jax.grad(qnode)(params)
    [Array(23.29654346, dtype=float64),
     Array(1.40236881, dtype=float64),
     Array(7.79013035, dtype=float64),
     Array(-4.33877283, dtype=float64),
     Array(-27.53388596, dtype=float64),
     Array(7.65313594, dtype=float64)]

    """
    if d != 2:
        raise NotImplementedError(
            "Currently only supports qubits (d=2). Qutrits and qudits support is planned in the future."
        )

    wires = Wires(wires)

    # TODO: use creation and annihilation operators when introducing qutrits
    # Note that exp(-iw)a* + exp(iw)a = cos(w)X - sin(w)Y for a=1/2(X+iY)
    # We compute the `coeffs` and `observables` of the EM field
    coeffs = [
        AmplitudeAndPhaseAndFreq(qml.math.cos, amplitude, phase, freq),
        AmplitudeAndPhaseAndFreq(qml.math.sin, amplitude, phase, freq),
    ]

    drive_x_term = sum(qml.PauliX(wire) for wire in wires)
    drive_y_term = sum(-qml.PauliY(wire) for wire in wires)

    observables = [drive_x_term, drive_y_term]

    pulses = [HardwarePulse(amplitude, phase, freq, wires)]

    return HardwareHamiltonian(coeffs, observables, pulses=pulses, reorder_fn=_reorder_AmpPhaseFreq)


# pylint:disable = too-few-public-methods,too-many-return-statements
class AmplitudeAndPhaseAndFreq:
    """Class storing combined amplitude, phase and freq callables"""

    def __init__(self, trig_fn, amp, phase, freq, hz_to_rads=2 * np.pi):
        self.amp_is_callable = callable(amp)
        self.phase_is_callable = callable(phase)
        self.freq_is_callable = callable(freq)

        # all 3 callable

        if self.amp_is_callable and self.phase_is_callable and self.freq_is_callable:

            def callable_amp_and_phase_and_freq(params, t):
                return (
                    hz_to_rads
                    * amp(params[0], t)
                    * trig_fn(phase(params[1], t) + hz_to_rads * freq(params[2], t) * t)
                )

            self.func = callable_amp_and_phase_and_freq
            return

        # 2 out of 3 callable

        if self.amp_is_callable and self.phase_is_callable:

            def callable_amp_and_phase(params, t):
                return (
                    hz_to_rads
                    * amp(params[0], t)
                    * trig_fn(phase(params[1], t) + hz_to_rads * freq * t)
                )

            self.func = callable_amp_and_phase
            return

        if self.amp_is_callable and self.freq_is_callable:

            def callable_amp_and_freq(params, t):
                return (
                    hz_to_rads
                    * amp(params[0], t)
                    * trig_fn(phase + hz_to_rads * freq(params[1], t) * t)
                )

            self.func = callable_amp_and_freq
            return

        if self.phase_is_callable and self.freq_is_callable:

            def callable_phase_and_freq(params, t):
                return (
                    hz_to_rads
                    * amp
                    * trig_fn(phase(params[0], t) + hz_to_rads * freq(params[1], t) * t)
                )

            self.func = callable_phase_and_freq
            return

        # 1 out of 3 callable

        if self.amp_is_callable:

            def callable_amp(params, t):
                return hz_to_rads * amp(params[0], t) * trig_fn(phase + hz_to_rads * freq * t)

            self.func = callable_amp
            return

        if self.phase_is_callable:

            def callable_phase(params, t):
                return hz_to_rads * amp * trig_fn(phase(params[0], t) + hz_to_rads * freq * t)

            self.func = callable_phase
            return

        if self.freq_is_callable:

            def callable_freq(params, t):
                return hz_to_rads * amp * trig_fn(phase + hz_to_rads * freq(params[0], t) * t)

            self.func = callable_freq
            return

        # 0 out of 3 callable
        # (the remaining coeff is still callable due to explicit time dependence)

        def no_callable(_, t):
            return hz_to_rads * amp * trig_fn(phase + hz_to_rads * freq * t)

        self.func = no_callable

    def __call__(self, params, t):
        return self.func(params, t)


def _reorder_AmpPhaseFreq(params, coeffs_parametrized):
    """Takes `params`, and reorganizes it based on whether the Hamiltonian has
    callable phase and/or callable amplitude and/or callable freq.

    Consolidates amplitude, phase and freq parameters in they are callable,
    and duplicates parameters since they will be passed to two operators in the Hamiltonian"""

    reordered_params = []

    coeff_idx = 0
    params_idx = 0

    for i, coeff in enumerate(coeffs_parametrized):
        if i == coeff_idx:
            if isinstance(coeff, AmplitudeAndPhaseAndFreq):
                is_callables = [
                    coeff.phase_is_callable,
                    coeff.amp_is_callable,
                    coeff.freq_is_callable,
                ]

                num_callables = sum(is_callables)

                # duplicate and package parameters according to how many coeffs are callable
                reordered_params.extend([params[params_idx : params_idx + num_callables]] * 2)

                coeff_idx += 2
                params_idx += num_callables

            else:
                reordered_params.append(params[params_idx])
                coeff_idx += 1
                params_idx += 1

    return reordered_params



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=too-many-arguments
"""
This file exists to support backwards compatibility for datasets pickled with
pennylane.ops.qubit.parametric_ops (i.e. from before the file splitting).

All new parametric operators should go into the more precisely named files.
"""

# pylint:disable=wildcard-import,unused-wildcard-import

from .parametric_ops_controlled import *
from .parametric_ops_multi_qubit import *
from .parametric_ops_single_qubit import *
